@@ -30,13 +30,16 @@
 #include <avi.h>
 #endif
 #else
-#include <ffmpeg/avformat.h>
+#include <libavformat/avformat.h>
+#ifdef HAVE_AVI_H
 #include <ffmpeg/avi.h>
 #endif
+#endif
 #include <gst/gst.h>
 
 #include "gstffmpeg.h"
 #include "gstffmpegcodecmap.h"
+#include "gstffmpegpipe.h"
 
 typedef struct _GstFFMpegDemux GstFFMpegDemux;
 typedef struct _GstFFStream GstFFStream;
@@ -71,8 +74,12 @@ struct _GstFFMpegDemux
   GstClockTime start_time;
   GstClockTime duration;
 
+  /* TRUE if working in pull-mode */
   gboolean seekable;
 
+  /* TRUE if the avformat demuxer can reliably handle streaming mode */
+  gboolean can_push;
+
   gboolean flushing;
 
   /* segment stuff */
@@ -81,6 +88,11 @@ struct _GstFFMpegDemux
 
   /* cached seek in READY */
   GstEvent *seek_event;
+
+  /* push mode data */
+  GstFFMpegPipe ffpipe;
+  GstTask *task;
+  GStaticRecMutex *task_lock;
 };
 
 typedef struct _GstFFMpegDemuxClassParams
@@ -105,8 +117,12 @@ struct _GstFFMpegDemuxClass
 static void gst_ffmpegdemux_class_init (GstFFMpegDemuxClass * klass);
 static void gst_ffmpegdemux_base_init (GstFFMpegDemuxClass * klass);
 static void gst_ffmpegdemux_init (GstFFMpegDemux * demux);
+static void gst_ffmpegdemux_finalize (GObject * object);
 
-static void gst_ffmpegdemux_loop (GstPad * pad);
+static gboolean gst_ffmpegdemux_sink_event (GstPad * sinkpad, GstEvent * event);
+static GstFlowReturn gst_ffmpegdemux_chain (GstPad * sinkpad, GstBuffer * buf);
+
+static void gst_ffmpegdemux_loop (GstFFMpegDemux * demux);
 static gboolean gst_ffmpegdemux_sink_activate (GstPad * sinkpad);
 static gboolean
 gst_ffmpegdemux_sink_activate_pull (GstPad * sinkpad, gboolean active);
@@ -143,9 +159,6 @@ gst_ffmpegdemux_averror (gint av_errno)
     case AVERROR_NUMEXPECTED:
       message = "Number syntax expected in filename";
       break;
-    case AVERROR_INVALIDDATA:
-      message = "Invalid data found";
-      break;
     case AVERROR_NOMEM:
       message = "Not enough memory";
       break;
@@ -189,6 +202,7 @@ gst_ffmpegdemux_base_init (GstFFMpegDemuxClass * klass)
   g_free (details.description);
 
   /* pad templates */
+  gst_caps_ref (params->sinkcaps);
   sinktempl = gst_pad_template_new ("sink",
       GST_PAD_SINK, GST_PAD_ALWAYS, params->sinkcaps);
   videosrctempl = gst_pad_template_new ("video_%02d",
@@ -196,6 +210,9 @@ gst_ffmpegdemux_base_init (GstFFMpegDemuxClass * klass)
   audiosrctempl = gst_pad_template_new ("audio_%02d",
       GST_PAD_SRC, GST_PAD_SOMETIMES, params->audiosrccaps);
 
+  params->videosrccaps = NULL;
+  params->audiosrccaps = NULL;
+
   gst_element_class_add_pad_template (element_class, videosrctempl);
   gst_element_class_add_pad_template (element_class, audiosrctempl);
   gst_element_class_add_pad_template (element_class, sinktempl);
@@ -217,6 +234,8 @@ gst_ffmpegdemux_class_init (GstFFMpegDemuxClass * klass)
 
   parent_class = g_type_class_peek_parent (klass);
 
+  gobject_class->finalize = GST_DEBUG_FUNCPTR (gst_ffmpegdemux_finalize);
+
   gstelement_class->change_state = gst_ffmpegdemux_change_state;
   gstelement_class->send_event = gst_ffmpegdemux_send_event;
 }
@@ -229,13 +248,26 @@ gst_ffmpegdemux_init (GstFFMpegDemux * demux)
   gint n;
 
   demux->sinkpad = gst_pad_new_from_template (oclass->sinktempl, "sink");
-  gst_pad_set_activate_function (demux->sinkpad, gst_ffmpegdemux_sink_activate);
+  gst_pad_set_activate_function (demux->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_ffmpegdemux_sink_activate));
   gst_pad_set_activatepull_function (demux->sinkpad,
-      gst_ffmpegdemux_sink_activate_pull);
+      GST_DEBUG_FUNCPTR (gst_ffmpegdemux_sink_activate_pull));
   gst_pad_set_activatepush_function (demux->sinkpad,
-      gst_ffmpegdemux_sink_activate_push);
+      GST_DEBUG_FUNCPTR (gst_ffmpegdemux_sink_activate_push));
   gst_element_add_pad (GST_ELEMENT (demux), demux->sinkpad);
 
+  /* push based setup */
+  /* the following are not used in pull-based mode, so safe to set anyway */
+  gst_pad_set_event_function (demux->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_ffmpegdemux_sink_event));
+  gst_pad_set_chain_function (demux->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_ffmpegdemux_chain));
+  /* task for driving ffmpeg in loop function */
+  demux->task = gst_task_create ((GstTaskFunction) gst_ffmpegdemux_loop, demux);
+  demux->task_lock = g_new (GStaticRecMutex, 1);
+  g_static_rec_mutex_init (demux->task_lock);
+  gst_task_set_lock (demux->task, demux->task_lock);
+
   demux->opened = FALSE;
   demux->context = NULL;
 
@@ -247,6 +279,35 @@ gst_ffmpegdemux_init (GstFFMpegDemux * demux)
 
   demux->seek_event = NULL;
   gst_segment_init (&demux->segment, GST_FORMAT_TIME);
+
+  /* push based data */
+  demux->ffpipe.tlock = g_mutex_new ();
+  demux->ffpipe.cond = g_cond_new ();
+  demux->ffpipe.adapter = gst_adapter_new ();
+
+  /* blacklist unreliable push-based demuxers */
+  if (strcmp (oclass->in_plugin->name, "ape"))
+    demux->can_push = TRUE;
+  else
+    demux->can_push = FALSE;
+}
+
+static void
+gst_ffmpegdemux_finalize (GObject * object)
+{
+  GstFFMpegDemux *demux;
+
+  demux = (GstFFMpegDemux *) object;
+
+  g_mutex_free (demux->ffpipe.tlock);
+  g_cond_free (demux->ffpipe.cond);
+  gst_object_unref (demux->ffpipe.adapter);
+
+  gst_object_unref (demux->task);
+  g_static_rec_mutex_free (demux->task_lock);
+  g_free (demux->task_lock);
+
+  G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 static void
@@ -336,6 +397,7 @@ gst_ffmpegdemux_is_eos (GstFFMpegDemux * demux)
 
   for (n = 0; n < MAX_STREAMS; n++) {
     if ((s = demux->streams[n])) {
+      GST_DEBUG ("stream %d %p eos:%d", n, s, s->eos);
       if (!s->eos)
         return FALSE;
     }
@@ -343,6 +405,22 @@ gst_ffmpegdemux_is_eos (GstFFMpegDemux * demux)
   return TRUE;
 }
 
+/* Returns True if we at least outputted one buffer */
+static gboolean
+gst_ffmpegdemux_has_outputted (GstFFMpegDemux * demux)
+{
+  GstFFStream *s;
+  gint n;
+
+  for (n = 0; n < MAX_STREAMS; n++) {
+    if ((s = demux->streams[n])) {
+      if (GST_CLOCK_TIME_IS_VALID (s->last_ts))
+        return TRUE;
+    }
+  }
+  return FALSE;
+}
+
 static gboolean
 gst_ffmpegdemux_do_seek (GstFFMpegDemux * demux, GstSegment * segment)
 {
@@ -433,6 +511,11 @@ gst_ffmpegdemux_perform_seek (GstFFMpegDemux * demux, GstEvent * event)
   gboolean update;
   GstSegment seeksegment;
 
+  if (!demux->seekable) {
+    GST_DEBUG_OBJECT (demux, "in push mode; ignoring seek");
+    return FALSE;
+  }
+
   GST_DEBUG_OBJECT (demux, "starting seek");
 
   if (event) {
@@ -614,6 +697,7 @@ gst_ffmpegdemux_src_query_list (GstPad * pad)
   static const GstQueryType src_types[] = {
     GST_QUERY_DURATION,
     GST_QUERY_POSITION,
+    GST_QUERY_SEEKING,
     0
   };
 
@@ -737,6 +821,22 @@ gst_ffmpegdemux_src_query (GstPad * pad, GstQuery * query)
       }
     }
       break;
+    case GST_QUERY_SEEKING:{
+      GstFormat format;
+      gboolean seekable;
+      gint64 dur = -1;
+
+      gst_query_parse_seeking (query, &format, NULL, NULL, NULL);
+      seekable = demux->seekable;
+      if (!gst_pad_query_duration (pad, &format, &dur)) {
+        /* unlikely that we don't know duration but can seek */
+        seekable = FALSE;
+        dur = -1;
+      }
+      gst_query_set_seeking (query, format, seekable, 0, dur);
+      res = TRUE;
+      break;
+    }
     default:
       /* FIXME : ADD GST_QUERY_CONVERT */
       res = gst_pad_query_default (pad, query);
@@ -895,7 +995,7 @@ gst_ffmpegdemux_get_stream (GstFFMpegDemux * demux, AVStream * avstream)
   stream->pad = pad;
   gst_pad_set_element_private (pad, stream);
 
-  /* transform some usefull info to GstClockTime and remember */
+  /* transform some useful info to GstClockTime and remember */
   {
     GstClockTime tmp;
 
@@ -1030,7 +1130,10 @@ gst_ffmpegdemux_open (GstFFMpegDemux * demux)
   gst_ffmpegdemux_close (demux);
 
   /* open via our input protocol hack */
-  location = g_strdup_printf ("gstreamer://%p", demux->sinkpad);
+  if (demux->seekable)
+    location = g_strdup_printf ("gstreamer://%p", demux->sinkpad);
+  else
+    location = g_strdup_printf ("gstpipe://%p", &demux->ffpipe);
   GST_DEBUG_OBJECT (demux, "about to call av_open_input_file %s", location);
 
   res = av_open_input_file (&demux->context, location,
@@ -1064,7 +1167,7 @@ gst_ffmpegdemux_open (GstFFMpegDemux * demux)
         gst_message_new_tag (GST_OBJECT (demux), tags));
   }
 
-  /* transform some usefull info to GstClockTime and remember */
+  /* transform some useful info to GstClockTime and remember */
   demux->start_time = gst_util_uint64_scale_int (demux->context->start_time,
       GST_SECOND, AV_TIME_BASE);
   GST_DEBUG_OBJECT (demux, "start time: %" GST_TIME_FORMAT,
@@ -1143,10 +1246,14 @@ gst_ffmpegdemux_type_find (GstTypeFind * tf, gpointer priv)
       res = MAX (1, res * GST_TYPE_FIND_MAXIMUM / AVPROBE_SCORE_MAX);
       /* Restrict the probability for MPEG-TS streams, because there is
        * probably a better version in plugins-base, if the user has a recent
-       * plugins-base */
-      if (!strcmp (in_plugin->name, "mpegts"))
+       * plugins-base (in fact we shouldn't even get here for ffmpeg mpegts or
+       * mpegtsraw typefinders, since we blacklist them) */
+      if (g_str_has_prefix (in_plugin->name, "mpegts"))
         res = MIN (res, GST_TYPE_FIND_POSSIBLE);
 
+      GST_LOG ("ffmpeg typefinder '%s' suggests %" GST_PTR_FORMAT ", p=%u%%",
+          in_plugin->name, params->sinkcaps, res);
+
       gst_type_find_suggest (tf, res, params->sinkcaps);
     }
   }
@@ -1154,19 +1261,18 @@ gst_ffmpegdemux_type_find (GstTypeFind * tf, gpointer priv)
 
 /* Task */
 static void
-gst_ffmpegdemux_loop (GstPad * pad)
+gst_ffmpegdemux_loop (GstFFMpegDemux * demux)
 {
-  GstFFMpegDemux *demux;
   GstFlowReturn ret;
   gint res;
   AVPacket pkt;
   GstPad *srcpad;
   GstFFStream *stream;
   AVStream *avstream;
-  GstBuffer *outbuf;
+  GstBuffer *outbuf = NULL;
   GstClockTime timestamp, duration;
-
-  demux = (GstFFMpegDemux *) (GST_PAD_PARENT (pad));
+  gint outsize;
+  gboolean rawvideo;
 
   /* open file if we didn't so already */
   if (!demux->opened)
@@ -1225,16 +1331,53 @@ gst_ffmpegdemux_loop (GstPad * pad)
   /* prepare to push packet to peer */
   srcpad = stream->pad;
 
-  ret = gst_pad_alloc_buffer_and_set_caps (srcpad,
-      GST_CLOCK_TIME_NONE, pkt.size, GST_PAD_CAPS (srcpad), &outbuf);
-  /* we can ignore not linked */
-  if (ret == GST_FLOW_NOT_LINKED)
-    goto done;
-  if (ret != GST_FLOW_OK)
+  rawvideo = (avstream->codec->codec_type == CODEC_TYPE_VIDEO &&
+      avstream->codec->codec_id == CODEC_ID_RAWVIDEO);
+
+  if (rawvideo)
+    outsize = gst_ffmpeg_avpicture_get_size (avstream->codec->pix_fmt,
+        avstream->codec->width, avstream->codec->height);
+  else
+    outsize = pkt.size;
+
+  stream->last_flow = gst_pad_alloc_buffer_and_set_caps (srcpad,
+      GST_CLOCK_TIME_NONE, outsize, GST_PAD_CAPS (srcpad), &outbuf);
+
+  if ((ret = gst_ffmpegdemux_aggregated_flow (demux)) != GST_FLOW_OK)
     goto no_buffer;
 
-  /* copy the data from packet into the target buffer */
-  memcpy (GST_BUFFER_DATA (outbuf), pkt.data, pkt.size);
+  /* If the buffer allocation failed, don't try sending it ! */
+  if (stream->last_flow != GST_FLOW_OK)
+    goto done;
+
+  /* copy the data from packet into the target buffer
+   * and do conversions for raw video packets */
+  if (rawvideo) {
+    AVPicture src, dst;
+    const gchar *plugin_name =
+        ((GstFFMpegDemuxClass *) (G_OBJECT_GET_CLASS (demux)))->in_plugin->name;
+
+    if (strcmp (plugin_name, "gif") == 0) {
+      src.data[0] = pkt.data;
+      src.data[1] = NULL;
+      src.data[2] = NULL;
+      src.linesize[0] = avstream->codec->width * 3;;
+    } else {
+      GST_WARNING ("Unknown demuxer %s, no idea what to do", plugin_name);
+      gst_ffmpeg_avpicture_fill (&src, pkt.data,
+          avstream->codec->pix_fmt, avstream->codec->width,
+          avstream->codec->height);
+    }
+
+    gst_ffmpeg_avpicture_fill (&dst, GST_BUFFER_DATA (outbuf),
+        avstream->codec->pix_fmt, avstream->codec->width,
+        avstream->codec->height);
+
+    av_picture_copy (&dst, &src, avstream->codec->pix_fmt,
+        avstream->codec->width, avstream->codec->height);
+  } else {
+    memcpy (GST_BUFFER_DATA (outbuf), pkt.data, outsize);
+  }
 
   GST_BUFFER_TIMESTAMP (outbuf) = timestamp;
   GST_BUFFER_DURATION (outbuf) = duration;
@@ -1277,28 +1420,42 @@ pause:
     GST_LOG_OBJECT (demux, "pausing task, reason %d (%s)", ret,
         gst_flow_get_name (ret));
     demux->running = FALSE;
-    gst_pad_pause_task (demux->sinkpad);
+    if (demux->seekable)
+      gst_pad_pause_task (demux->sinkpad);
+    else {
+      GstFFMpegPipe *ffpipe = &demux->ffpipe;
 
-    if (ret == GST_FLOW_UNEXPECTED) {
-      if (demux->segment.flags & GST_SEEK_FLAG_SEGMENT) {
-        gint64 stop;
+      GST_FFMPEG_PIPE_MUTEX_LOCK (ffpipe);
+      /* pause task and make sure loop stops */
+      gst_task_pause (demux->task);
+      g_static_rec_mutex_lock (demux->task_lock);
+      g_static_rec_mutex_unlock (demux->task_lock);
+      demux->ffpipe.srcresult = ret;
+      GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+    }
 
-        if ((stop = demux->segment.stop) == -1)
-          stop = demux->segment.duration;
+    if (GST_FLOW_IS_FATAL (ret) || ret == GST_FLOW_NOT_LINKED) {
+      if (ret == GST_FLOW_UNEXPECTED) {
+        if (demux->segment.flags & GST_SEEK_FLAG_SEGMENT) {
+          gint64 stop;
 
-        GST_LOG_OBJECT (demux, "posting segment done");
-        gst_element_post_message (GST_ELEMENT (demux),
-            gst_message_new_segment_done (GST_OBJECT (demux),
-                demux->segment.format, stop));
+          if ((stop = demux->segment.stop) == -1)
+            stop = demux->segment.duration;
+
+          GST_LOG_OBJECT (demux, "posting segment done");
+          gst_element_post_message (GST_ELEMENT (demux),
+              gst_message_new_segment_done (GST_OBJECT (demux),
+                  demux->segment.format, stop));
+        } else {
+          GST_LOG_OBJECT (demux, "pushing eos");
+          gst_ffmpegdemux_push_event (demux, gst_event_new_eos ());
+        }
       } else {
-        GST_LOG_OBJECT (demux, "pushing eos");
+        GST_ELEMENT_ERROR (demux, STREAM, FAILED,
+            ("Internal data stream error."),
+            ("streaming stopped, reason %s", gst_flow_get_name (ret)));
         gst_ffmpegdemux_push_event (demux, gst_event_new_eos ());
       }
-    } else if (GST_FLOW_IS_FATAL (ret)) {
-      GST_ELEMENT_ERROR (demux, STREAM, FAILED,
-          ("Internal data stream error."),
-          ("streaming stopped, reason %s", gst_flow_get_name (ret)));
-      gst_ffmpegdemux_push_event (demux, gst_event_new_eos ());
     }
     return;
   }
@@ -1316,8 +1473,12 @@ read_failed:
     /* pause appropriatly based on if we are flushing or not */
     if (demux->flushing)
       ret = GST_FLOW_WRONG_STATE;
-    else
+    else if (gst_ffmpegdemux_has_outputted (demux)
+        || gst_ffmpegdemux_is_eos (demux)) {
+      GST_DEBUG_OBJECT (demux, "We are EOS");
       ret = GST_FLOW_UNEXPECTED;
+    } else
+      ret = GST_FLOW_ERROR;
     GST_OBJECT_UNLOCK (demux);
 
     goto pause;
@@ -1344,6 +1505,135 @@ no_buffer:
 }
 
 
+static gboolean
+gst_ffmpegdemux_sink_event (GstPad * sinkpad, GstEvent * event)
+{
+  GstFFMpegDemux *demux;
+  GstFFMpegPipe *ffpipe;
+  gboolean result = TRUE;
+
+  demux = (GstFFMpegDemux *) (GST_PAD_PARENT (sinkpad));
+  ffpipe = &(demux->ffpipe);
+
+  GST_DEBUG_OBJECT (demux, "event %s", GST_EVENT_TYPE_NAME (event));
+
+  switch (GST_EVENT_TYPE (event)) {
+    case GST_EVENT_FLUSH_START:
+      /* forward event */
+      gst_pad_event_default (sinkpad, event);
+
+      /* now unblock the chain function */
+      GST_FFMPEG_PIPE_MUTEX_LOCK (ffpipe);
+      ffpipe->srcresult = GST_FLOW_WRONG_STATE;
+      GST_FFMPEG_PIPE_SIGNAL (ffpipe);
+      GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+
+      /* loop might run into WRONG_STATE and end itself,
+       * but may also be waiting in a ffmpeg read
+       * trying to break that would make ffmpeg believe eos,
+       * so no harm to have the loop 'pausing' there ... */
+      goto done;
+    case GST_EVENT_FLUSH_STOP:
+      /* forward event */
+      gst_pad_event_default (sinkpad, event);
+
+      GST_FFMPEG_PIPE_MUTEX_LOCK (ffpipe);
+      gst_adapter_clear (ffpipe->adapter);
+      ffpipe->srcresult = GST_FLOW_OK;
+      /* loop may have decided to end itself as a result of flush WRONG_STATE */
+      gst_task_start (demux->task);
+      demux->running = TRUE;
+      demux->flushing = FALSE;
+      GST_LOG_OBJECT (demux, "loop started");
+      GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+      goto done;
+    case GST_EVENT_EOS:
+      /* inform the src task that it can stop now */
+      GST_FFMPEG_PIPE_MUTEX_LOCK (ffpipe);
+      ffpipe->eos = TRUE;
+      GST_FFMPEG_PIPE_SIGNAL (ffpipe);
+      GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+
+      /* eat this event for now, task will send eos when finished */
+      gst_event_unref (event);
+      goto done;
+    default:
+      /* for a serialized event, wait until an earlier data is gone,
+       * though this is no guarantee as to when task is done with it.
+       *
+       * If the demuxer isn't opened, push straight away, since we'll
+       * be waiting against a cond that will never be signalled. */
+      if (GST_EVENT_IS_SERIALIZED (event) && demux->opened) {
+        GST_FFMPEG_PIPE_MUTEX_LOCK (ffpipe);
+        while (!ffpipe->needed)
+          GST_FFMPEG_PIPE_WAIT (ffpipe);
+        GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+      }
+      break;
+  }
+
+  result = gst_pad_event_default (sinkpad, event);
+
+done:
+
+  return result;
+}
+
+static GstFlowReturn
+gst_ffmpegdemux_chain (GstPad * sinkpad, GstBuffer * buffer)
+{
+  GstFFMpegDemux *demux;
+  GstFFMpegPipe *ffpipe;
+
+  demux = (GstFFMpegDemux *) (GST_PAD_PARENT (sinkpad));
+  ffpipe = &demux->ffpipe;
+
+  GST_FFMPEG_PIPE_MUTEX_LOCK (ffpipe);
+
+  if (G_UNLIKELY (ffpipe->eos))
+    goto eos;
+
+  if (G_UNLIKELY (ffpipe->srcresult != GST_FLOW_OK))
+    goto ignore;
+
+  GST_DEBUG ("Giving a buffer of %d bytes", GST_BUFFER_SIZE (buffer));
+  gst_adapter_push (ffpipe->adapter, buffer);
+  buffer = NULL;
+  while (gst_adapter_available (ffpipe->adapter) >= ffpipe->needed) {
+    GST_DEBUG ("Adapter has more that requested (ffpipe->needed:%d)",
+        ffpipe->needed);
+    GST_FFMPEG_PIPE_SIGNAL (ffpipe);
+    GST_FFMPEG_PIPE_WAIT (ffpipe);
+    /* may have become flushing */
+    if (G_UNLIKELY (ffpipe->srcresult != GST_FLOW_OK))
+      goto ignore;
+  }
+
+  GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+
+  return GST_FLOW_OK;
+
+/* special cases */
+eos:
+  {
+    GST_DEBUG_OBJECT (demux, "ignoring buffer at end-of-stream");
+    GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+
+    gst_buffer_unref (buffer);
+    return GST_FLOW_UNEXPECTED;
+  }
+ignore:
+  {
+    GST_DEBUG_OBJECT (demux, "ignoring buffer because src task encountered %s",
+        gst_flow_get_name (ffpipe->srcresult));
+    GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+
+    if (buffer)
+      gst_buffer_unref (buffer);
+    return GST_FLOW_WRONG_STATE;
+  }
+}
+
 static gboolean
 gst_ffmpegdemux_sink_activate (GstPad * sinkpad)
 {
@@ -1363,23 +1653,61 @@ gst_ffmpegdemux_sink_activate (GstPad * sinkpad)
   return res;
 }
 
+/* push mode:
+ * - not seekable
+ * - use gstpipe protocol, like ffmpeg's pipe protocol
+ * - (independently managed) task driving ffmpeg
+ */
 static gboolean
 gst_ffmpegdemux_sink_activate_push (GstPad * sinkpad, gboolean active)
 {
   GstFFMpegDemux *demux;
+  gboolean res = FALSE;
 
   demux = (GstFFMpegDemux *) (gst_pad_get_parent (sinkpad));
 
-  GST_ELEMENT_ERROR (demux, STREAM, NOT_IMPLEMENTED,
-      (NULL),
-      ("failed to activate sinkpad in pull mode, push mode not implemented yet"));
+  if (active) {
+    if (demux->can_push == FALSE) {
+      GST_WARNING_OBJECT (demux, "Demuxer can't reliably operate in push-mode");
+      goto beach;
+    }
+    demux->ffpipe.eos = FALSE;
+    demux->ffpipe.srcresult = GST_FLOW_OK;
+    demux->ffpipe.needed = 0;
+    demux->running = TRUE;
+    demux->seekable = FALSE;
+    res = gst_task_start (demux->task);
+  } else {
+    GstFFMpegPipe *ffpipe = &demux->ffpipe;
 
-  demux->seekable = FALSE;
+    /* release chain and loop */
+    GST_FFMPEG_PIPE_MUTEX_LOCK (ffpipe);
+    demux->ffpipe.srcresult = GST_FLOW_WRONG_STATE;
+    /* end streaming by making ffmpeg believe eos */
+    demux->ffpipe.eos = TRUE;
+    GST_FFMPEG_PIPE_SIGNAL (ffpipe);
+    GST_FFMPEG_PIPE_MUTEX_UNLOCK (ffpipe);
+
+    /* make sure streaming ends */
+    gst_task_stop (demux->task);
+    g_static_rec_mutex_lock (demux->task_lock);
+    g_static_rec_mutex_unlock (demux->task_lock);
+    res = gst_task_join (demux->task);
+    demux->running = FALSE;
+    demux->seekable = FALSE;
+  }
+
+beach:
   gst_object_unref (demux);
 
-  return FALSE;
+  return res;
 }
 
+/* pull mode:
+ * - seekable
+ * - use gstreamer protocol, like ffmpeg's file protocol
+ * - task driving ffmpeg based on sink pad
+ */
 static gboolean
 gst_ffmpegdemux_sink_activate_pull (GstPad * sinkpad, gboolean active)
 {
@@ -1392,7 +1720,7 @@ gst_ffmpegdemux_sink_activate_pull (GstPad * sinkpad, gboolean active)
     demux->running = TRUE;
     demux->seekable = TRUE;
     res = gst_pad_start_task (sinkpad, (GstTaskFunction) gst_ffmpegdemux_loop,
-        sinkpad);
+        demux);
   } else {
     demux->running = FALSE;
     res = gst_pad_stop_task (sinkpad);
@@ -1429,6 +1757,7 @@ gst_ffmpegdemux_change_state (GstElement * element, GstStateChange transition)
   switch (transition) {
     case GST_STATE_CHANGE_PAUSED_TO_READY:
       gst_ffmpegdemux_close (demux);
+      gst_adapter_clear (demux->ffpipe.adapter);
       break;
     default:
       break;
@@ -1458,6 +1787,8 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
 
   in_plugin = first_iformat;
 
+  GST_LOG ("Registering demuxers");
+
   while (in_plugin) {
     gchar *type_name, *typefind_name;
     gchar *p, *name = NULL;
@@ -1465,6 +1796,9 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     gint rank;
     gboolean register_typefind_func = TRUE;
 
+    GST_LOG ("Attempting to handle ffmpeg demuxer plugin %s [%s]",
+        in_plugin->name, in_plugin->long_name);
+
     /* no emulators */
     if (!strncmp (in_plugin->long_name, "raw ", 4) ||
         !strncmp (in_plugin->long_name, "pcm ", 4) ||
@@ -1472,7 +1806,19 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
         !strncmp (in_plugin->name, "image", 5) ||
         !strcmp (in_plugin->name, "mpegvideo") ||
         !strcmp (in_plugin->name, "mjpeg") ||
-        !strcmp (in_plugin->name, "redir"))
+        !strcmp (in_plugin->name, "redir") ||
+        !strncmp (in_plugin->name, "u8", 2) ||
+        !strncmp (in_plugin->name, "u16", 3) ||
+        !strncmp (in_plugin->name, "u24", 3) ||
+        !strncmp (in_plugin->name, "u32", 3) ||
+        !strncmp (in_plugin->name, "s8", 2) ||
+        !strncmp (in_plugin->name, "s16", 3) ||
+        !strncmp (in_plugin->name, "s24", 3) ||
+        !strncmp (in_plugin->name, "s32", 3) ||
+        !strncmp (in_plugin->name, "f32", 3) ||
+        !strncmp (in_plugin->name, "f64", 3) ||
+        !strcmp (in_plugin->name, "mulaw") || !strcmp (in_plugin->name, "alaw")
+        )
       goto next;
 
     /* no network demuxers */
@@ -1481,12 +1827,13 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
 
     /* these don't do what one would expect or
      * are only partially functional/useful */
-    if (!strcmp (in_plugin->name, "aac"))
+    if (!strcmp (in_plugin->name, "aac") || !strcmp (in_plugin->name, "wv"))
       goto next;
 
     /* Don't use the typefind functions of formats for which we already have
      * better typefind functions */
     if (!strcmp (in_plugin->name, "mov,mp4,m4a,3gp,3g2,mj2") ||
+        !strcmp (in_plugin->name, "ass") ||
         !strcmp (in_plugin->name, "avi") ||
         !strcmp (in_plugin->name, "asf") ||
         !strcmp (in_plugin->name, "mpegvideo") ||
@@ -1496,14 +1843,29 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
         !strcmp (in_plugin->name, "wav") ||
         !strcmp (in_plugin->name, "au") ||
         !strcmp (in_plugin->name, "tta") ||
-        !strcmp (in_plugin->name, "rm") || !strcmp (in_plugin->name, "amr"))
+        !strcmp (in_plugin->name, "rm") ||
+        !strcmp (in_plugin->name, "amr") ||
+        !strcmp (in_plugin->name, "ogg") ||
+        !strcmp (in_plugin->name, "aiff") ||
+        !strcmp (in_plugin->name, "ape") ||
+        !strcmp (in_plugin->name, "dv") ||
+        !strcmp (in_plugin->name, "flv") ||
+        !strcmp (in_plugin->name, "mpc") ||
+        !strcmp (in_plugin->name, "mpc8") ||
+        !strcmp (in_plugin->name, "mpegts") ||
+        !strcmp (in_plugin->name, "mpegtsraw") ||
+        !strcmp (in_plugin->name, "mxf") ||
+        !strcmp (in_plugin->name, "nuv") ||
+        !strcmp (in_plugin->name, "swf") ||
+        !strcmp (in_plugin->name, "voc") ||
+        !strcmp (in_plugin->name, "gif") || !strcmp (in_plugin->name, "vc1test")
+        )
       register_typefind_func = FALSE;
 
     /* Set the rank of demuxers know to work to MARGINAL.
      * Set demuxers for which we already have another implementation to NONE
      * Set All others to NONE*/
-    if (!strcmp (in_plugin->name, "flv") ||
-        !strcmp (in_plugin->name, "wsvqa") ||
+    if (!strcmp (in_plugin->name, "wsvqa") ||
         !strcmp (in_plugin->name, "wsaud") ||
         !strcmp (in_plugin->name, "wc3movie") ||
         !strcmp (in_plugin->name, "voc") ||
@@ -1522,6 +1884,7 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
         !strcmp (in_plugin->name, "mmf") ||
         !strcmp (in_plugin->name, "mm") ||
         !strcmp (in_plugin->name, "ipmovie") ||
+        !strcmp (in_plugin->name, "ape") ||
         !strcmp (in_plugin->name, "RoQ") ||
         !strcmp (in_plugin->name, "idcin") ||
         !strcmp (in_plugin->name, "gxf") ||
@@ -1531,7 +1894,8 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
         !strcmp (in_plugin->name, "avs") ||
         !strcmp (in_plugin->name, "aiff") ||
         !strcmp (in_plugin->name, "4xm") ||
-        !strcmp (in_plugin->name, "yuv4mpegpipe"))
+        !strcmp (in_plugin->name, "yuv4mpegpipe") ||
+        !strcmp (in_plugin->name, "mpc") || !strcmp (in_plugin->name, "gif"))
       rank = GST_RANK_MARGINAL;
     else
       rank = GST_RANK_NONE;
@@ -1546,6 +1910,7 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     /* Try to find the caps that belongs here */
     sinkcaps = gst_ffmpeg_formatid_to_caps (name);
     if (!sinkcaps) {
+      GST_WARNING ("Couldn't get sinkcaps for demuxer %s", in_plugin->name);
       goto next;
     }
     /* This is a bit ugly, but we just take all formats
@@ -1604,5 +1969,7 @@ gst_ffmpegdemux_register (GstPlugin * plugin)
     in_plugin = in_plugin->next;
   }
 
+  GST_LOG ("Finished registering demuxers");
+
   return TRUE;
 }
