@@ -25,7 +25,7 @@
 #ifdef HAVE_FFMPEG_UNINSTALLED
 #include <avformat.h>
 #else
-#include <ffmpeg/avformat.h>
+#include <libavformat/avformat.h>
 #endif
 
 #include <gst/gst.h>
@@ -55,13 +55,13 @@ struct _GstFFMpegMux
   AVFormatContext *context;
   gboolean opened;
 
-  GstTagList *tags;
-
   gint videopads, audiopads;
 
   /*< private > */
   /* event_function is the collectpads default eventfunction */
   GstPadEventFunction event_function;
+  int preload;
+  int max_delay;
 };
 
 typedef struct _GstFFMpegMuxClassParams
@@ -102,6 +102,13 @@ enum
   /* FILL ME */
 };
 
+enum
+{
+  PROP_0,
+  PROP_PRELOAD,
+  PROP_MAXDELAY
+};
+
 /* A number of function prototypes are given so we can refer to them later. */
 static void gst_ffmpegmux_class_init (GstFFMpegMuxClass * klass);
 static void gst_ffmpegmux_base_init (gpointer g_class);
@@ -120,6 +127,11 @@ static gboolean gst_ffmpegmux_sink_event (GstPad * pad, GstEvent * event);
 static GstStateChangeReturn gst_ffmpegmux_change_state (GstElement * element,
     GstStateChange transition);
 
+static void gst_ffmpegmux_set_property (GObject * object, guint prop_id,
+    const GValue * value, GParamSpec * pspec);
+static void gst_ffmpegmux_get_property (GObject * object, guint prop_id,
+    GValue * value, GParamSpec * pspec);
+
 #define GST_FFMUX_PARAMS_QDATA g_quark_from_static_string("ffmux-params")
 
 static GstElementClass *parent_class = NULL;
@@ -184,6 +196,19 @@ gst_ffmpegmux_class_init (GstFFMpegMuxClass * klass)
 
   parent_class = g_type_class_peek_parent (klass);
 
+  gobject_class->set_property = GST_DEBUG_FUNCPTR (gst_ffmpegmux_set_property);
+  gobject_class->get_property = GST_DEBUG_FUNCPTR (gst_ffmpegmux_get_property);
+
+  g_object_class_install_property (gobject_class, PROP_PRELOAD,
+      g_param_spec_int ("preload", "preload",
+          "Set the initial demux-decode delay (in microseconds)", 0, G_MAXINT,
+          0, G_PARAM_READWRITE));
+
+  g_object_class_install_property (gobject_class, PROP_MAXDELAY,
+      g_param_spec_int ("maxdelay", "maxdelay",
+          "Set the maximum demux-decode delay (in microseconds)", 0, G_MAXINT,
+          0, G_PARAM_READWRITE));
+
   gstelement_class->request_new_pad = gst_ffmpegmux_request_new_pad;
   gstelement_class->change_state = gst_ffmpegmux_change_state;
   gobject_class->finalize = gst_ffmpegmux_finalize;
@@ -213,10 +238,53 @@ gst_ffmpegmux_init (GstFFMpegMux * ffmpegmux, GstFFMpegMuxClass * g_class)
 
   ffmpegmux->videopads = 0;
   ffmpegmux->audiopads = 0;
-
-  ffmpegmux->tags = NULL;
+  ffmpegmux->preload = 0;
+  ffmpegmux->max_delay = 0;
 }
 
+static void
+gst_ffmpegmux_set_property (GObject * object, guint prop_id,
+    const GValue * value, GParamSpec * pspec)
+{
+  GstFFMpegMux *src;
+
+  src = (GstFFMpegMux *) object;
+
+  switch (prop_id) {
+    case PROP_PRELOAD:
+      src->preload = g_value_get_int (value);
+      break;
+    case PROP_MAXDELAY:
+      src->max_delay = g_value_get_int (value);
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+  }
+}
+
+static void
+gst_ffmpegmux_get_property (GObject * object, guint prop_id, GValue * value,
+    GParamSpec * pspec)
+{
+  GstFFMpegMux *src;
+
+  src = (GstFFMpegMux *) object;
+
+  switch (prop_id) {
+    case PROP_PRELOAD:
+      g_value_set_int (value, src->preload);
+      break;
+    case PROP_MAXDELAY:
+      g_value_set_int (value, src->max_delay);
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+  }
+}
+
+
 static void
 gst_ffmpegmux_finalize (GObject * object)
 {
@@ -313,16 +381,27 @@ gst_ffmpegmux_setcaps (GstPad * pad, GstCaps * caps)
   collect_pad = (GstFFMpegMuxPad *) gst_pad_get_element_private (pad);
 
   st = ffmpegmux->context->streams[collect_pad->padnum];
+  ffmpegmux->context->preload = ffmpegmux->preload;
+  ffmpegmux->context->max_delay = ffmpegmux->max_delay;
 
   /* for the format-specific guesses, we'll go to
    * our famous codec mapper */
-  if (gst_ffmpeg_caps_to_codecid (caps, st->codec) != CODEC_ID_NONE) {
-    GST_LOG_OBJECT (pad, "accepted caps %" GST_PTR_FORMAT, caps);
-    return TRUE;
-  }
+  if (gst_ffmpeg_caps_to_codecid (caps, st->codec) == CODEC_ID_NONE)
+    goto not_accepted;
 
-  GST_LOG_OBJECT (pad, "rejecting caps %" GST_PTR_FORMAT, caps);
-  return FALSE;
+  /* copy over the aspect ratios, ffmpeg expects the stream aspect to match the
+   * codec aspect. */
+  st->sample_aspect_ratio = st->codec->sample_aspect_ratio;
+
+  GST_LOG_OBJECT (pad, "accepted caps %" GST_PTR_FORMAT, caps);
+  return TRUE;
+
+  /* ERRORS */
+not_accepted:
+  {
+    GST_LOG_OBJECT (pad, "rejecting caps %" GST_PTR_FORMAT, caps);
+    return FALSE;
+  }
 }
 
 
@@ -334,12 +413,12 @@ gst_ffmpegmux_sink_event (GstPad * pad, GstEvent * event)
 
   switch (GST_EVENT_TYPE (event)) {
     case GST_EVENT_TAG:{
-      GstTagList *taglist = NULL;
+      GstTagList *taglist;
+      GstTagSetter *setter = GST_TAG_SETTER (ffmpegmux);
+      const GstTagMergeMode mode = gst_tag_setter_get_tag_merge_mode (setter);
 
       gst_event_parse_tag (event, &taglist);
-      ffmpegmux->tags = gst_tag_list_merge (ffmpegmux->tags, taglist,
-          GST_TAG_MERGE_PREPEND);
-
+      gst_tag_setter_merge_tags (setter, taglist, mode);
       break;
     }
     default:
@@ -360,7 +439,7 @@ gst_ffmpegmux_collected (GstCollectPads * pads, gpointer user_data)
   GSList *collected;
   GstFFMpegMuxPad *best_pad;
   GstClockTime best_time;
-  const GstTagList *iface_tags;
+  const GstTagList *tags;
 
   /* open "file" (gstreamer protocol to next element) */
   if (!ffmpegmux->opened) {
@@ -412,15 +491,11 @@ gst_ffmpegmux_collected (GstCollectPads * pads, gpointer user_data)
     }
 
     /* tags */
-    iface_tags = gst_tag_setter_get_tag_list (GST_TAG_SETTER (ffmpegmux));
-    if (ffmpegmux->tags || iface_tags) {
-      GstTagList *tags;
+    tags = gst_tag_setter_get_tag_list (GST_TAG_SETTER (ffmpegmux));
+    if (tags) {
       gint i;
       gchar *s;
 
-      tags = gst_tag_list_merge (iface_tags, ffmpegmux->tags,
-          GST_TAG_MERGE_APPEND);
-
       /* get the interesting ones */
       if (gst_tag_list_get_string (tags, GST_TAG_TITLE, &s)) {
         strncpy (ffmpegmux->context->title, s,
@@ -449,7 +524,6 @@ gst_ffmpegmux_collected (GstCollectPads * pads, gpointer user_data)
       if (gst_tag_list_get_int (tags, GST_TAG_TRACK_NUMBER, &i)) {
         ffmpegmux->context->track = i;
       }
-      gst_tag_list_free (tags);
     }
 
     /* set the streamheader flag for gstffmpegprotocol if codec supports it */
@@ -481,7 +555,7 @@ gst_ffmpegmux_collected (GstCollectPads * pads, gpointer user_data)
     ffmpegmux->opened = TRUE;
 
     /* flush the header so it will be used as streamheader */
-    put_flush_packet (&ffmpegmux->context->pb);
+    put_flush_packet (ffmpegmux->context->pb);
   }
 
   /* take the one with earliest timestamp,
@@ -525,6 +599,7 @@ gst_ffmpegmux_collected (GstCollectPads * pads, gpointer user_data)
   if (best_pad != NULL) {
     GstBuffer *buf;
     AVPacket pkt;
+    gboolean need_free = FALSE;
 
     /* push out current buffer */
     buf = gst_collect_pads_pop (ffmpegmux->collect,
@@ -536,8 +611,30 @@ gst_ffmpegmux_collected (GstCollectPads * pads, gpointer user_data)
     pkt.pts = gst_ffmpeg_time_gst_to_ff (GST_BUFFER_TIMESTAMP (buf),
         ffmpegmux->context->streams[best_pad->padnum]->time_base);
     pkt.dts = pkt.pts;
-    pkt.data = GST_BUFFER_DATA (buf);
-    pkt.size = GST_BUFFER_SIZE (buf);
+
+    if (strcmp (ffmpegmux->context->oformat->name, "gif") == 0) {
+      AVStream *st = ffmpegmux->context->streams[best_pad->padnum];
+      AVPicture src, dst;
+
+      need_free = TRUE;
+      pkt.size = st->codec->width * st->codec->height * 3;
+      pkt.data = g_malloc (pkt.size);
+
+      dst.data[0] = pkt.data;
+      dst.data[1] = NULL;
+      dst.data[2] = NULL;
+      dst.linesize[0] = st->codec->width * 3;
+
+      gst_ffmpeg_avpicture_fill (&src, GST_BUFFER_DATA (buf),
+          PIX_FMT_RGB24, st->codec->width, st->codec->height);
+
+      av_picture_copy (&dst, &src, PIX_FMT_RGB24,
+          st->codec->width, st->codec->height);
+    } else {
+      pkt.data = GST_BUFFER_DATA (buf);
+      pkt.size = GST_BUFFER_SIZE (buf);
+    }
+
     pkt.stream_index = best_pad->padnum;
     pkt.flags = 0;
 
@@ -546,18 +643,20 @@ gst_ffmpegmux_collected (GstCollectPads * pads, gpointer user_data)
 
     if (GST_BUFFER_DURATION_IS_VALID (buf))
       pkt.duration =
-          gst_util_uint64_scale_int (GST_BUFFER_DURATION (buf), AV_TIME_BASE,
-          GST_SECOND);
+          gst_ffmpeg_time_gst_to_ff (GST_BUFFER_DURATION (buf),
+          ffmpegmux->context->streams[best_pad->padnum]->time_base);
     else
       pkt.duration = 0;
     av_write_frame (ffmpegmux->context, &pkt);
     gst_buffer_unref (buf);
+    if (need_free)
+      g_free (pkt.data);
   } else {
     /* close down */
     av_write_trailer (ffmpegmux->context);
     ffmpegmux->opened = FALSE;
-    put_flush_packet (&ffmpegmux->context->pb);
-    url_fclose (&ffmpegmux->context->pb);
+    put_flush_packet (ffmpegmux->context->pb);
+    url_fclose (ffmpegmux->context->pb);
     gst_pad_push_event (ffmpegmux->srcpad, gst_event_new_eos ());
     return GST_FLOW_UNEXPECTED;
   }
@@ -592,13 +691,10 @@ gst_ffmpegmux_change_state (GstElement * element, GstStateChange transition)
     case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
       break;
     case GST_STATE_CHANGE_PAUSED_TO_READY:
-      if (ffmpegmux->tags) {
-        gst_tag_list_free (ffmpegmux->tags);
-        ffmpegmux->tags = NULL;
-      }
+      gst_tag_setter_reset_tags (GST_TAG_SETTER (ffmpegmux));
       if (ffmpegmux->opened) {
         ffmpegmux->opened = FALSE;
-        url_fclose (&ffmpegmux->context->pb);
+        url_fclose (ffmpegmux->context->pb);
       }
       break;
     case GST_STATE_CHANGE_READY_TO_NULL:
@@ -673,20 +769,50 @@ gst_ffmpegmux_register (GstPlugin * plugin)
 
   in_plugin = first_oformat;
 
+  GST_LOG ("Registering muxers");
+
   while (in_plugin) {
     gchar *type_name;
     gchar *p;
     GstCaps *srccaps, *audiosinkcaps, *videosinkcaps;
     enum CodecID *video_ids = NULL, *audio_ids = NULL;
 
+    if ((!strncmp (in_plugin->name, "u16", 3)) ||
+        (!strncmp (in_plugin->name, "s16", 3)) ||
+        (!strncmp (in_plugin->name, "u24", 3)) ||
+        (!strncmp (in_plugin->name, "s24", 3)) ||
+        (!strncmp (in_plugin->name, "u8", 2)) ||
+        (!strncmp (in_plugin->name, "s8", 2)) ||
+        (!strncmp (in_plugin->name, "u32", 3)) ||
+        (!strncmp (in_plugin->name, "s32", 3)) ||
+        (!strncmp (in_plugin->name, "f32", 3)) ||
+        (!strncmp (in_plugin->name, "f64", 3)) ||
+        (!strncmp (in_plugin->name, "raw", 3)) ||
+        (!strncmp (in_plugin->name, "crc", 3)) ||
+        (!strncmp (in_plugin->name, "null", 4)) ||
+        (!strncmp (in_plugin->name, "gif", 3)) ||
+        (!strncmp (in_plugin->name, "frame", 5)) ||
+        (!strncmp (in_plugin->name, "image", 5)) ||
+        (!strncmp (in_plugin->name, "mulaw", 5)) ||
+        (!strncmp (in_plugin->name, "alaw", 4)) ||
+        (!strncmp (in_plugin->name, "h26", 3))
+        ) {
+      GST_LOG ("Ignoring muxer %s", in_plugin->name);
+      goto next;
+    }
+
     /* Try to find the caps that belongs here */
     srccaps = gst_ffmpeg_formatid_to_caps (in_plugin->name);
     if (!srccaps) {
+      GST_WARNING ("Couldn't get source caps for muxer %s", in_plugin->name);
       goto next;
     }
     if (!gst_ffmpeg_formatid_get_codecids (in_plugin->name,
-            &video_ids, &audio_ids)) {
+            &video_ids, &audio_ids, in_plugin)) {
       gst_caps_unref (srccaps);
+      GST_WARNING
+          ("Couldn't get sink caps for muxer %s. Most likely because no input format mapping exists.",
+          in_plugin->name);
       goto next;
     }
     videosinkcaps = video_ids ? gst_ffmpegmux_get_id_caps (video_ids) : NULL;
@@ -719,6 +845,12 @@ gst_ffmpegmux_register (GstPlugin * plugin)
       const gint rates[] = { 44100, 22050, 11025 };
 
       gst_ffmpeg_mux_simple_caps_set_int_list (audiosinkcaps, "rate", 3, rates);
+    } else if (strcmp (in_plugin->name, "gif") == 0) {
+      if (videosinkcaps)
+        gst_caps_unref (videosinkcaps);
+
+      videosinkcaps =
+          gst_caps_from_string ("video/x-raw-rgb, bpp=(int)24, depth=(int)24");
     }
 
     /* create a cache for these properties */
@@ -749,5 +881,7 @@ gst_ffmpegmux_register (GstPlugin * plugin)
     in_plugin = in_plugin->next;
   }
 
+  GST_LOG ("Finished registering muxers");
+
   return TRUE;
 }
